from unittest.mock import NonCallableMagicMock, call

from mltome.neptune import NeptuneSkorchCallback


def test_neptunecallback():
    ctx_mock = NonCallableMagicMock()
    no = NeptuneSkorchCallback(
        ctx_mock, batch_targets=['train_loss'], epoch_targets=['valid_loss'])

    no.update_batch_values({'train_loss': 1}, 0)
    ctx_mock.channel_send.assert_has_calls([call('train_loss', 0, 1)])

    no.update_epoch_values({'valid_loss': 1}, 0)
    ctx_mock.channel_send.assert_has_calls([call('valid_loss', 0, 1)])



from .skorch.callbacks import MetricsLogger


class CometSkorchCallback(MetricsLogger):
    def __init__(self, exp, batch_targets=None, epoch_targets=None):
        super().__init__(
            batch_targets=batch_targets, epoch_targets=epoch_targets)
        self.exp = exp

    def update_batch_values(self, values, idx):
        self.exp.log_metrics(values, step=idx)

    def update_epoch_values(self, values, idx):
        self.exp.log_metrics(values, step=idx)



"""Skorch callbacks"""
from itertools import product
from collections import defaultdict
from contextlib import suppress

from skorch.callbacks.base import Callback


class MetricsLogger(Callback):
    """Base class for creating metrics loggers. Child classes
    must implement ``update_batch_values`` and ``update_epoch_values``
    to use this logger

    Parameters
    ----------

    batch_targets : list of str
        Name of metrics to log at the end of each batch

    epoch_targets : list of str
        Name of metrics to log at the end of each epoch

    """

    def __init__(self, batch_targets=None, epoch_targets=None):

        self.batch_targets = batch_targets
        self.epoch_targets = epoch_targets

    def initialize(self):
        self.batch_idx_ = 0

    def update_batch_values(self, values, idx):
        raise NotImplementedError("Not implemented")

    def update_epoch_values(self, values, idx):
        raise NotImplementedError("Not implemented")

    def on_batch_end(self, net, **kwargs):
        if self.batch_targets is None:
            return

        values = {}
        for name in self.batch_targets:
            with suppress(KeyError):
                values[name] = net.history[-1, 'batches', -1, name]

        self.update_batch_values(values, self.batch_idx_)
        self.batch_idx_ += 1

    def on_epoch_end(self, net, **kwargs):
        if self.epoch_targets is None:
            return

        epoch = len(net.history)

        values = {}
        for name in self.epoch_targets:
            values[name] = net.history[-1, name]

        self.update_epoch_values(values, epoch)


class LRRecorder(Callback):
    """Records learning rate from optimizer

    Parameters
    ----------

    group_names : list of str
        Name of each parameter group

    per_epoch : bool
        If true, record learning rate for each epoch, otherwise
        record each batch

    default_group : str
        Name of default learning rate group

    """

    def __init__(self,
                 group_names=None,
                 per_epoch=True,
                 default_group="default_lr"):
        self.group_names = group_names
        self.per_epoch = per_epoch
        self.default_group = default_group

    def initialize(self):
        group_names = self.group_names
        if group_names is None:
            group_names = []
        if self.default_group:
            group_names.append(self.default_group)
        self.group_names_ = group_names

    def on_train_begin(self, net, **kwargs):
        self.optimizer_ = net.optimizer_

    def on_epoch_end(self, net, **kwargs):
        if not self.per_epoch:
            return
        history = net.history
        pgroups = self.optimizer_.param_groups

        for pgroup, name in zip(pgroups, self.group_names_):
            history.record(name, pgroup['lr'])

    def on_batch_end(self, net, **kwargs):
        if self.per_epoch:
            return
        history = net.history
        pgroups = self.optimizer_.param_groups

        for pgroup, name in zip(pgroups, self.group_names_):
            history.record_batch(name, pgroup['lr'])


class TensorboardXLogger(MetricsLogger):
    """TensorboardX callbacks

    Parameters
    ----------

    name : str
        Name of current run

    batch_targets : list of str
        Targets to record per batch

    epoch_targets : list of str
        Targets to record per epoch

    batch_groups : list of str
        Groups to record per batch

    epoch_groupss : list of str
        Groups to record per epoch

    """

    def __init__(self,
                 log_dir,
                 batch_targets=None,
                 epoch_targets=None,
                 batch_groups=None,
                 epoch_groups=None):
        self.log_dir = log_dir

        self.batch_groups = batch_groups or []
        self.epoch_groups = epoch_groups or []
        self.batch_targets = batch_targets or []
        self.epoch_targets = epoch_targets or []

        super().__init__(
            batch_targets=batch_targets, epoch_targets=epoch_targets)

    def initialize(self):
        super().initialize()
        from tensorboardX import SummaryWriter

        self.writer_ = SummaryWriter(log_dir=self.log_dir)
        self.batch_target_to_name_ = {}
        for g, t in product(self.batch_groups, self.batch_targets):
            if t.endswith(g):
                self.batch_target_to_name_[t] = 'batch_' + g

        self.epoch_target_to_name_ = {}
        for g, t in product(self.epoch_groups, self.epoch_targets):
            if t.endswith(g):
                self.epoch_target_to_name_[t] = 'epoch_' + g

    def update_batch_values(self, values, idx):
        vgroups = defaultdict(dict)
        for name, value in values.items():
            self.writer_.add_scalar(f'batch/{name}', value, idx)
            with suppress(KeyError):
                group = self.batch_target_to_name_[name]
                vgroups[group][name] = value

        for group, values in vgroups.items():
            self.writer_.add_scalars(group, values, idx)

    def update_epoch_values(self, values, idx):
        vgroups = defaultdict(dict)
        for name, value in values.items():
            self.writer_.add_scalar(f'epoch/{name}', value, idx)
            with suppress(KeyError):
                group = self.epoch_target_to_name_[name]
                vgroups[group][name] = value

        for group, values in vgroups.items():
            self.writer_.add_scalars(group, values, idx)

    def on_train_end(self, net, **kwargs):
        self.writer_.close()



from contextlib import suppress

import numpy as np
from skorch.callbacks.base import Callback

from .callbacks import LRRecorder


class LRFinder(Callback):
    def __init__(self, start_lr, end_lr, warm_start=10, scale_linear=False):

        self.start_lrs = start_lr
        self.end_lrs = end_lr
        self.scale_linear = scale_linear

        self.lr_multiplier = None
        self.warm_start = warm_start
        self.final_valid_batch_idx = None

    def initialize(self):
        self.batch_idx_ = 0

    def on_train_begin(self, net, X, **kwargs):
        self.best_loss = 1e9
        self.total_samples = net.max_epochs * len(X)
        self.best_batch_idx = 0

        optimizer = net.optimizer_
        self.start_lrs = self._format_lrs("start_lr", optimizer,
                                          self.start_lrs)
        self.end_lrs = self._format_lrs("end_lr", optimizer, self.end_lrs)
        self.optimizer = optimizer

    def on_batch_begin(self, net, X, **kwargs):
        if self.lr_multiplier is None:
            ratio = self.end_lrs / self.start_lrs
            num_of_batches = self.total_samples // len(X)

            self.lr_multiplier = ratio / num_of_batches
            if not self.scale_linear:
                self.lr_multiplier = ratio**(1 / num_of_batches)

        self.batch_step(self.batch_idx_)

    def on_batch_end(self, net, **kwargs):
        loss = net.history[-1, 'batches', -1, 'train_loss']
        if not np.isfinite(loss) or loss > 4 * self.best_loss:
            raise ValueError("loss is too big")
        if loss < self.best_loss and self.batch_idx_ > self.warm_start:
            self.best_loss = loss
            self.best_lr = self.get_lr(self.batch_idx_)
        self.batch_idx_ += 1

    def _format_lrs(self, name, optimizer, lr):
        if isinstance(lr, (list, tuple)):
            if len(lr) != len(optimizer.param_groups):
                raise ValueError(
                    f"expected {len(optimizer.param_groups)} values for "
                    f"{name}, got {len(lr)}")
            return np.array(lr)
        else:
            return lr * np.ones(len(optimizer.param_groups))

    def batch_step(self, batch_idx):
        lrs = self.get_lr(batch_idx)
        pgroups_lr = zip(self.optimizer.param_groups, lrs)
        for param_group, lr in pgroups_lr:
            param_group['lr'] = lr

    def get_lr(self, batch_idx):
        mult = self.lr_multiplier * (batch_idx + 1)
        if not self.scale_linear:
            mult = self.lr_multiplier**(batch_idx + 1)
        return self.start_lrs * mult


def lr_find(net_cls,
            module,
            criterion,
            batch_size,
            X,
            iterations,
            y=None,
            start_lr=1e-5,
            end_lr=10,
            scale_linear=False,
            **kwargs):
    lr_finder = ('lr_finder',
                 LRFinder(
                     start_lr=start_lr,
                     end_lr=end_lr,
                     scale_linear=scale_linear))
    lr_recorder = ('lr_recorder', LRRecorder(per_epoch=False))
    callbacks = [lr_finder, lr_recorder]

    epochs = iterations // len(X)

    net = net_cls(
        module,
        criterion=criterion,
        max_epochs=epochs,
        batch_size=batch_size,
        callbacks=callbacks,
        train_split=None,
        **kwargs)

    with suppress(ValueError):
        net.fit(X, y)

    return net, lr_finder[1]


def plot_lr(history, lr_finder, ax):
    if not lr_finder.scale_linear:
        ax.set_xscale('log')

    losses = history[:, 'batches', 'train_loss'][0]
    lrs = history[:, 'batches', 'default_lr'][0]
    ax.vlines(lr_finder.best_lr, min(losses), max(losses), color='r')
    ax.set_ylim(min(losses), max(losses))
    ax.set_xlabel("Learning rate (log-scaled)")
    ax.set_ylabel("Loss")
    ax.plot(lrs, losses)



from unittest.mock import ANY, call, Mock

import pytest
import numpy as np
import torch
import torch.nn as nn

from skorch.net import NeuralNet
from mltome.skorch.callbacks import LRRecorder
from mltome.skorch.callbacks import TensorboardXLogger


class MyModule(nn.Module):
    def __init__(self):
        super().__init__()
        self.layer1 = nn.Linear(1, 2)
        self.layer2 = nn.Linear(2, 1)

    def forward(self, x):
        return self.layer2(self.layer1(x))


@pytest.fixture
def data():
    X = np.ones((1000, 1), dtype='float32')
    y = np.ones((1000, 1), dtype='float32')
    return X, y


# LRRecorder
def test_lr_recorder_default_lr_per_epoch(data):
    module = MyModule()
    X, y = data

    recorder = LRRecorder(per_epoch=True)
    net = NeuralNet(
        module,
        torch.nn.MSELoss,
        callbacks=[recorder],
        max_epochs=5,
        optimizer__lr=0.1,
        batch_size=100)
    net.fit(X, y)

    lrs = net.history[:, 'default_lr']
    assert lrs == [0.1] * 5


def test_lr_recorder_default_lr_per_batch(data):
    module = MyModule()
    X, y = data

    recorder = LRRecorder(per_epoch=False)
    net = NeuralNet(
        module,
        torch.nn.MSELoss,
        callbacks=[recorder],
        max_epochs=5,
        optimizer__lr=0.1,
        batch_size=100)
    net.fit(X, y)

    lrs_per_epoch = net.history[:, 'batches', :, 'default_lr']
    for lr in lrs_per_epoch:
        assert lr == [0.1] * 10


def test_lr_recorder_custom_lr_per_epoch(data):
    module = MyModule()
    X, y = data

    param_group = [('layer1.*', {'lr': 0.01})]
    params_names = [name for name, _ in param_group]
    recorder = LRRecorder(group_names=params_names, per_epoch=True)

    net = NeuralNet(
        module,
        torch.nn.MSELoss,
        callbacks=[recorder],
        max_epochs=5,
        optimizer__lr=0.1,
        batch_size=100,
        optimizer__param_groups=param_group)
    net.fit(X, y)

    lrs = net.history[:, 'layer1.*']
    assert lrs == [0.01] * 5

    lrs = net.history[:, 'default_lr']
    assert lrs == [0.10] * 5


def test_lr_recorder_custom_lr_per_batch(data):
    module = MyModule()
    X, y = data

    param_group = [('layer1.*', {'lr': 0.01})]
    params_names = [name for name, _ in param_group]
    recorder = LRRecorder(group_names=params_names, per_epoch=False)

    net = NeuralNet(
        module,
        torch.nn.MSELoss,
        callbacks=[recorder],
        max_epochs=5,
        optimizer__lr=0.1,
        batch_size=100,
        optimizer__param_groups=param_group)
    net.fit(X, y)

    lrs_per_epoch_layer1 = net.history[:, 'batches', :, 'layer1.*']
    for lr in lrs_per_epoch_layer1:
        assert lr == [0.01] * 10

    lrs_per_epoch_default = net.history[:, 'batches', :, 'default_lr']
    for lr in lrs_per_epoch_default:
        assert lr == [0.1] * 10


def test_tensorboard_logger(monkeypatch, data, tmpdir):
    add_scalar_mock = Mock()
    add_scalars_mock = Mock()

    monkeypatch.setattr('tensorboardX.SummaryWriter.add_scalar',
                        add_scalar_mock)
    monkeypatch.setattr('tensorboardX.SummaryWriter.add_scalars',
                        add_scalars_mock)

    log_dir = str(tmpdir.mkdir('tensorboardx_log'))

    module = MyModule()

    tensor_logger = TensorboardXLogger(
        log_dir,
        batch_targets=['train_loss'],
        epoch_targets=['valid_loss'],
        batch_groups=['train_loss'],
        epoch_groups=['valid_loss'])
    net = NeuralNet(
        module,
        torch.nn.MSELoss,
        callbacks=[tensor_logger],
        max_epochs=2,
        batch_size=100)

    X, y = data

    net.fit(X, y)

    batch_calls = [call('batch/train_loss', ANY, i) for i in range(8)]
    batch_calls += [call('batch/train_loss', ANY, i) for i in range(10, 18)]
    epoch_calls = [call('epoch/valid_loss', ANY, i) for i in range(1, 3)]
    add_scalar_mock.assert_has_calls(batch_calls + epoch_calls, any_order=True)

    batch_group_calls = [call('batch_train_loss', ANY, i) for i in range(8)]
    batch_group_calls += [
        call('batch_train_loss', ANY, i) for i in range(10, 18)
    ]
    epoch_group_calls = [call('epoch_valid_loss', ANY, i) for i in range(1, 3)]
    add_scalars_mock.assert_has_calls(
        batch_group_calls + epoch_group_calls, any_order=True)



import pytest
import numpy as np
import torch.nn as nn
from skorch import NeuralNetRegressor

from mltome.skorch.lr_finder import lr_find


class MyModule(nn.Module):
    def __init__(self):
        super().__init__()
        self.layer1 = nn.Linear(1, 2)
        self.layer2 = nn.Linear(2, 1)

    def forward(self, x):
        return self.layer2(self.layer1(x))


@pytest.fixture
def data():
    X = np.ones((1000, 1), dtype='float32')
    y = np.ones((1000, 1), dtype='float32')
    return X, y


def test_lr_find(data):
    X, y = data
    net, lr_finder = lr_find(NeuralNetRegressor, MyModule, nn.MSELoss, 100, X,
                             1000, y)

    assert net.max_epochs == 1



from .skorch.callbacks import MetricsLogger


class NeptuneSkorchCallback(MetricsLogger):
    def __init__(self, ctx, batch_targets=None, epoch_targets=None):
        super().__init__(
            batch_targets=batch_targets, epoch_targets=epoch_targets)
        self.ctx = ctx

    def update_batch_values(self, values, idx):
        for name, value in values.items():
            self.ctx.channel_send(name, idx, value)

    def update_epoch_values(self, values, idx):
        for name, value in values.items():
            self.ctx.channel_send(name, idx, value)



#!/usr/bin/env python

import os

from codecs import open
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open("requirements/base.txt") as f:
    install_requires = f.read().strip().split("\n")

with open("requirements/dev.txt") as f:
    dev_requires = f.read().strip().split("\n")

with open("VERSION", "r") as f:
    version = f.read().rstrip()

with open("README.rst", "r", "utf-8") as f:
    readme = f.read()

setup(
    name="mltome",
    version=version,
    description="A collection of Mtools that builds on Python ML frameworks.",
    long_description=readme,
    author="Thomas Fan",
    author_email="thomasjpfan@gmail.com",
    url="https://github.com/thomasjpfan/mltome",
    packages=find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    python_requires=">=3.5",
    zip_safe=False,
    license="MIT",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
    ],
    extras_require={"dev": dev_requires},
)
